'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _extends = Object.assign || function (target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i]; for (var key in source) { if (Object.prototype.hasOwnProperty.call(source, key)) { target[key] = source[key]; } } } return target; };

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _utils = require('../utils');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

function _possibleConstructorReturn(self, call) { if (!self) { throw new ReferenceError("this hasn't been initialised - super() hasn't been called"); } return call && (typeof call === "object" || typeof call === "function") ? call : self; }

function _inherits(subClass, superClass) { if (typeof superClass !== "function" && superClass !== null) { throw new TypeError("Super expression must either be null or a function, not " + typeof superClass); } subClass.prototype = Object.create(superClass && superClass.prototype, { constructor: { value: subClass, enumerable: false, writable: true, configurable: true } }); if (superClass) Object.setPrototypeOf ? Object.setPrototypeOf(subClass, superClass) : subClass.__proto__ = superClass; }

var Canvas = function (_React$Component) {
  _inherits(Canvas, _React$Component);

  function Canvas(props) {
    _classCallCheck(this, Canvas);

    var _this = _possibleConstructorReturn(this, Object.getPrototypeOf(Canvas).call(this, props));

    _this.getBitmapData = _this.getBitmapData.bind(_this);
    _this.renderCanvas = _this.renderCanvas.bind(_this);
    // private properties
    _this._prevAngle = 0;
    _this._data = null;
    _this._cache = null;
    return _this;
  }

  /**
   * Constrain proportions to rect while considering its angle
   * @param from {{width: number, height: number}}
   * @param to {{width: number, height: number}}
   * @returns {{width: number, height: number}}
   */

  _createClass(Canvas, [{
    key: 'getBitmapData',
    value: function getBitmapData(source) {
      try {
        if (source !== this._data) {
          this._cache = new Image();
          this._data = Object.assign(source);
          this._cache.src = this._data;
        }
        return this._cache;
      } catch (er) {
        this._cache = null;
        this._data = null;
        this._cache.src = null;
        return null;
      }
    }
  }, {
    key: 'renderCanvas',
    value: function renderCanvas() {
      var _this2 = this;

      var canvas = document.createElement('canvas'),
          ctx = canvas.getContext('2d');

      canvas.width = this.props.width;
      canvas.height = this.props.height;

      if (this.cache) {
        canvas = this.refs.canvas;
        ctx = canvas.getContext('2d');

        _utils.Transform.clearCanvas(canvas, ctx);

        ctx.drawImage(this.cache, 0, 0);
        if (this.props.crop) {
          if (!this.cropBox) {
            this.cropBox = new _utils.CropBox({ canvas: canvas, ctx: ctx });
          }
          this.cropBox.render();
        }
      }

      if (this.props.source && !this.cache) {

        var image = this.getBitmapData(this.props.source),
            angle = parseInt(this.props.angle) || 0,
            boundRect = {
          width: parseInt(this.props.width),
          height: parseInt(this.props.height)
        },
            dims = {
          width: parseInt(image.width),
          height: parseInt(image.height)
        };

        _utils.Transform.rotateImage(ctx, angle);

        var scaledRect = _utils.Transform.constrainProportions(dims, boundRect);
        var position = _utils.Transform.centerRect(scaledRect, boundRect);

        _utils.Transform.renderImage(ctx, image, position, scaledRect);
        setTimeout(function () {
          var img = new Image();
          img.src = canvas.toDataURL("image/png");
          _this2.cache = img;
        }, 100);
      }
    }
  }, {
    key: 'componentDidMount',
    value: function componentDidMount() {
      var _this3 = this;

      var canvas = this.refs.canvas,
          ctx = canvas.getContext('2d');

      setInterval(function () {
        _this3.renderCanvas();
      }, 30);
    }
  }, {
    key: 'componentWillUpdate',
    value: function componentWillUpdate() {
      this.cache = null;
    }
  }, {
    key: 'render',
    value: function render() {
      var _props = this.props;
      var source = _props.source;
      var width = _props.width;
      var height = _props.height;
      var children = _props.children;
      var crop = _props.crop;


      var canvasStyle = {
        display: source ? 'block' : 'none'
      };

      return _react2.default.createElement(
        'div',
        _extends({}, this.props, { style: { width: width, height: height }, className: 'darkroom-canvas' }),
        children,
        _react2.default.createElement('canvas', { style: canvasStyle, ref: 'canvas', width: width, height: height })
      );
    }
  }]);

  return Canvas;
}(_react2.default.Component);

exports.default = Canvas;